import pysvn


NODE_TYPE_NONE = 'N'
NODE_TYPE_FILE = 'F'
NODE_TYPE_DIR = 'D'
NODE_TYPE_UNKNOWN = '?'

NODE_TYPES = (
    (NODE_TYPE_NONE, 'Absent'),
    (NODE_TYPE_FILE, 'File'),
    (NODE_TYPE_DIR, 'Dir'),
    (NODE_TYPE_UNKNOWN, 'Unknown'),
)

NODE_TYPE_MAP = {
    pysvn.node_kind.none: NODE_TYPE_NONE,
    pysvn.node_kind.file: NODE_TYPE_FILE,
    pysvn.node_kind.dir: NODE_TYPE_DIR,
    pysvn.node_kind.unknown: NODE_TYPE_UNKNOWN,
}



import pysvn
from django import forms
from django.contrib import admin

from svnlit import models


class RepositoryForm(forms.ModelForm):
    class Meta:
        model = models.Repository

    password = forms.CharField(
        max_length=512, required=False, widget=forms.PasswordInput)


class RepositoryAdmin(admin.ModelAdmin):
    form = RepositoryForm

    fieldsets = (
        (None, {'fields': ('label', 'root', 'uri', 'is_private')}),
        (u'Credentials', {'fields': ('username', 'password')}),
    )


admin.site.register(models.Repository, RepositoryAdmin)



import pysvn


class SubversionError(Exception): pass

class InvalidNode(SubversionError): pass
class RepositoryNotFound(SubversionError): pass


def map_svn_exceptions(func):
    def new_func(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except pysvn.ClientError, error:
            error_code = error.args[1][0][1]
            if error_code in svn_error_map:
                raise svn_error_map[error_code]
            else:
                raise error

    new_func.__name__ = func.__name__
    new_func.__dict__ = func.__dict__
    new_func.__doc__ = func.__doc__
    new_func.__module__ = func.__module__

    return new_func


# map of common svn errors encountered by svnlit
svn_error_map = {
    160013: InvalidNode,
    210005: RepositoryNotFound,
}



from django.conf.urls.defaults import *


urlpatterns = patterns(
    'svnlit.views',

    (r'^$',
     'repository_list', None, 'svnlit_repository_list'),

    (r'^(?P<repository_label>[\w-]+):c(?P<content_id>[0-9]+)/(?P<path>.+)$',
     'content', None, 'svnlit_content'),

    (r'^(?P<repository_label>[\w-]+)$',
     'changeset_list', None, 'svnlit_changeset_list'),

    (r'^(?P<repository_label>[\w-]+):diff$',
     'repository_diff', None, 'svnlit_repository_diff'),

    (r'^(?P<repository_label>[\w-]+):r(?P<revision>[0-9]+)$',
     'changeset', None, 'svnlit_changeset'),

    (r'^(?P<repository_label>[\w-]+):r(?P<from_revision>[0-9]+)-r(?P<to_revision>[0-9]+)(?P<path>/.*)$',
     'node_diff', None, 'svnlit_node_diff'),

    (r'^(?P<repository_label>[\w-]+):r(?P<revision>[0-9]+)(?P<path>/.*)$',
     'node', None, 'svnlit_node_revision'),

    (r'^(?P<repository_label>[\w-]+)(?P<path>/.*)$',
     'node', {'revision': None}, 'svnlit_node')
)



from django import forms
from django.utils.translation import ugettext_lazy as _

from svnlit import models, exceptions


class RepositoryDiffForm(forms.Form):
    from_path = forms.CharField(
        max_length=1024, widget=forms.TextInput(attrs={'size': '40'}))
    from_revision = forms.CharField(
        widget=forms.TextInput(attrs={'size': '3'}))
    to_path = forms.CharField(
        max_length=1024, widget=forms.TextInput(attrs={'size': '40'}))
    to_revision = forms.CharField(
        widget=forms.TextInput(attrs={'size': '3'}))

    def is_valid(self, repository):
        self.repository = repository
        return super(RepositoryDiffForm, self).is_valid()

    def clean(self):
        repository = self.repository
        from_path = self.cleaned_data.get('from_path', '')
        from_revision = self.cleaned_data.get('from_revision', '')
        to_path = self.cleaned_data.get('to_path', '')
        to_revision = self.cleaned_data.get('to_revision', '')

        if not (from_path and from_revision and to_path and to_revision):
            raise forms.ValidationError(
                _(('Please specify both path and '
                   'revision combinations to compare.')))

        if not from_revision.isdigit() or not to_revision.isdigit():
            raise forms.ValidationError(_('Revision should be a number.'))
        from_revision, to_revision = int(from_revision), int(to_revision)

        try:
            from_node = repository.get_node(from_path, from_revision)
        except exceptions.InvalidNode:
            raise forms.ValidationError(
                _('Non-existant node at the path/revision comparing from.'))
        try:
            to_node = repository.get_node(to_path, to_revision)
        except exceptions.InvalidNode:
            raise forms.ValidationError(
                _('Non-existant node at the path/revision comparing to.'))

        return {'from_node': from_node, 'to_node': to_node}



import datetime

from django.conf import settings

from svnlit import models


_last_sync_check = datetime.datetime.fromtimestamp(0)


def autosync_repositories(view_func):
    def new_view_func(request, *args, **kwargs):
        global _last_sync_check

        if settings.SVNLIT_AUTO_SYNC:
            sync_td = datetime.timedelta(0, settings.SVNLIT_SYNC_INTERVAL)
            if datetime.datetime.now() - _last_sync_check > sync_td:
                threshold = datetime.datetime.now() - sync_td
                outdated_repositories = models.Repository.objects.filter(
                    last_synced__lte=threshold)
                for repository in outdated_repositories:
                    repository.sync()
                _last_sync_check = datetime.datetime.now()
                
        return view_func(request, *args, **kwargs)

    new_view_func.__name__ = view_func.__name__
    new_view_func.__dict__ = view_func.__dict__
    new_view_func.__doc__ = view_func.__doc__
    new_view_func.__module__ = view_func.__module__

    return new_view_func



import datetime

from django.conf import settings
from django.core import management
from django.utils.translation import ugettext as _

from svnlit import models


class Command(management.BaseCommand):
    help = _('Clean up expired cached files')
    
    def handle(self, *args, **options):
        threshold = datetime.datetime.now() - datetime.timedelta(
            0, settings.SVNLIT_CACHE_TIMEOUT)
        nodes = models.Node.objects.filter(cached__lte=threshold)
        if nodes.count():
            print _('Deleting %s nodes...') % nodes.count()
            nodes.delete()



"""
Model managers that interface with the subversion API.
"""

import datetime
import posixpath

import pysvn
from django.conf import settings
from django.db.models import manager
from django.db import transaction

from svnlit import choices, exceptions


class RepositoryManager(manager.Manager):
    def get_svn_client(self, repository):
        """
        Return a subversion client for this repository. The
        authentication information stored with the repository is bound
        with the client. The client can be instantiated with a
        subversion config file with the SVNLIT_SVN_CONFIG_PATH
        settings variable.
        """
        if settings.SVNLIT_SVN_CONFIG_PATH is not None:
            client = pysvn.Client(settings.SVNLIT_SVN_CONFIG_PATH)
        else:
            client = pysvn.Client()

        # set the exceptions to be more granular
        client.exception_style = 1

        # hook for cancelling an api call thats taking too long
        started_dt = datetime.datetime.now()
        def _cancel():
            current_dt = datetime.datetime.now()
            delta = (current_dt - started_dt).seconds
            if delta > settings.SVNLIT_CLIENT_TIMEOUT:
                return True
            return False
        client.callback_cancel = _cancel

        # bind the username and password that might be stored with the
        # repository model object in case a login is required.
        client.callback_get_login = repository._get_login

        return client

    def sync(self, repository):
        """
        Update the model object representations of the given repository.

        If the UUID has not been obtained for a repository, it is
        obtained from the api. New changesets committed to the
        repository, since the last time the repository was synced, are
        also collected. If no previous sync has been run, all
        changesets are collected.
        """
        repository.last_synced = datetime.datetime.now()

        if not repository.uuid:
            self.sync_uuid(repository)
        self.sync_changesets(repository)

        repository.save()

    def sync_uuid(self, repository):
        """Get the UUID of the given repository."""
        c = self.get_svn_client(repository)
        info = c.info2(repository.root, recurse=False)
        repository.uuid = info[0][1]['repos_UUID']
    sync_uuid = exceptions.map_svn_exceptions(sync_uuid)
    
    def sync_changesets(self, repository):
        """
        Get new changesets committed to the repository since the last
        time they were collected.
        """
        from svnlit import models

        revision = repository.get_latest_revision()

        c = self.get_svn_client(repository)
        log = c.log(
            repository.root,
            revision_end=pysvn.Revision(
            pysvn.opt_revision_kind.number, revision),
            discover_changed_paths=True)

        for item in log:
            # ignore the overlap, the changeset is already stored locally
            if item['revision'].number == revision:
                continue

            changeset = models.Changeset.objects.create(
                repository=repository,
                date=datetime.datetime.fromtimestamp(item['date']),
                revision=item['revision'].number,
                author=item['author'],
                message=item['message'])

            for changed_path in item['changed_paths']:
                copyfrom_revision = None
                if changed_path['copyfrom_revision']:
                    copyfrom_revision = changed_path[
                        'copyfrom_revision'].number
                
                change = models.Change.objects.create(
                    changeset=changeset,
                    path=changed_path['path'],
                    action=changed_path['action'],
                    copied_from_path=changed_path['copyfrom_path'],
                    copied_from_revision=copyfrom_revision)

    sync_changesets = transaction.commit_on_success(
        exceptions.map_svn_exceptions(sync_changesets))


class NodeManager(manager.Manager):
    def get_or_sync(self, repository, path, revision=None):
        """
        Get a node in the given repository at the given path.

        Optionally specify a revision number. If the node has already
        been cached locally, the model object is returned. If it has
        not yet been cached, the node information is retrieved from an
        api call and cached locally before returning it.

        If the node is a file, and the contents have not been
        retrieved, the node's content is also obtained from the api.
        """
        from svnlit import models

        if revision is None:
            revision = repository.get_latest_revision()
        # removing the trailing slash from the path
        if path.endswith(posixpath.sep) and len(path) > 1:
            path = path[:-1]
        # add slash to begining of path
        if not path.startswith(posixpath.sep):
            path = '/%s' % path

        try:
            node = self.get(
                repository=repository, revision=revision, path=path)

            # return a directory node if its been fully cached
            if (node.is_directory() and
                not node.cached_indirectly):
                return node

            # return a file node if its been fully cached and it's
            # content has been retrieved
            if (node.is_file() and
                not node.cached_indirectly
                and node.content is not None):
                return node

        except self.model.DoesNotExist:
            # no cached node was found, fetch it from the api
            pass

        # construct a full uri to the requested node
        if repository.root.endswith(posixpath.sep):
            root = repository.root[:-1]
        else:
            root = repository.root
        root_path = '%s%s' % (root, path)

        c = models.Repository.objects.get_svn_client(repository)
        r = pysvn.Revision(pysvn.opt_revision_kind.number, revision)

        # get a listing of nodes for the given path. File nodes are
        # returned as a list containing one node, and directory nodes
        # are returned as a list with the first node in the list being
        # the directory node specified, followed by all its child
        # nodes.
        ls = c.list(root_path, recurse=False, peg_revision=r, revision=r)
        ls = map(lambda y: dict(y.items()), map(lambda x: x[0], ls))

        node_list = list()
        for item in ls:
            # sometimes double slashes appear in the returned path
            node_path = item['repos_path']
            if node_path.startswith('//'):
                node_path = node_path[1:]

            # when dealing with directory nodes, some or all child
            # nodes may already have been cached as model objects
            # locally.
            try:
                node = self.get(
                    repository=repository, path=node_path, revision=revision)
            except self.model.DoesNotExist:
                node = self.model(
                    repository=repository, path=node_path, revision=revision)

            node.node_type = choices.NODE_TYPE_MAP[item['kind']]
            node.last_changed = datetime.datetime.fromtimestamp(item['time'])
            node.size = item['size']
            node.cached_datetime = datetime.datetime.now()
            node_list.append(node)

        # separate out the primary node from any child nodes following it.
        # the primary node represents the repository/path/revision passed
        # to the get_or_sync method
        primary_node = filter(lambda n: n.path == path, node_list)[0]
        node_list.remove(primary_node)

        # the primary node is now fully cached
        primary_node.cached_indirectly = False

        # if the primary node is a directory, set all the child nodes'
        # parent to the primary node to establish the hierarchy within
        # the local model objects
        if primary_node.is_directory():
            for node in node_list:
                if node.path.startswith(primary_node.path):
                    node.parent = primary_node

        # if the primary node is a file, retrieve the content of the node.
        if primary_node.is_file() and primary_node.content is None:
            try:
                # applicable content for the primary node's
                # path/revision may already be cached as a model
                # object.
                primary_node.content = models.Content.objects.get(
                    repository=repository, path=primary_node.path,
                    last_changed=primary_node.last_changed)
            except models.Content.DoesNotExist:
                # otherwise the content is fetched from an api call
                content = models.Content(
                    repository=repository, path=primary_node.path,
                    last_changed=primary_node.last_changed)
                content.set_data(c.cat(root_path, revision=r, peg_revision=r))
                content.save()
                primary_node.content = content

        # obtain the parent node (if its not '/') if it hasn't been
        # obtained yet.
        if not primary_node.parent and not primary_node.is_root():
            parent_path = posixpath.dirname(primary_node.path)
            try:
                # the parent node may already be cached as a model object
                parent_node = self.get(
                    repository=repository, path=parent_path, revision=revision)
            except self.model.DoesNotExist:
                # create an indirectly cached node to use as the parent
                # (an additional api call will be needed later on to fill out the
                # rest of the information)
                parent_node = self.create(
                    repository=repository, path=parent_path, revision=revision,
                    node_type=choices.NODE_TYPE_DIR)
            primary_node.parent = parent_node

        primary_node.save()
        for n in node_list:
            n.parent = primary_node
            n.save()

        # collect the subversion properties set on the primary node
        # TODO: not sure if flushing is necessary here. It looks like
        # properties set at a given path/revision can be changed, so
        # we need to take care to check for updates.
        primary_node.properties.all().delete()
        prop_list = c.proplist(root_path, revision=r, peg_revision=r)
        for prop in prop_list:
            for key, value in prop[1].items():
                models.Property.objects.create(
                    node=primary_node, key=key, value=value)

        return primary_node

    get_or_sync = transaction.commit_on_success(
        exceptions.map_svn_exceptions(get_or_sync))



import re
import difflib


DIFF_OPTCODE_MAP = {' ': 'nil', '-': 'sub', '+': 'add', '^': 'chg'}


def optcodes_to_slices(optcodes):
    if optcodes.startswith('? '):
        optcodes = optcodes[2:-1]
    sections = filter(lambda x: x, re.split('( +)', optcodes))
    slices = []
    for x in range(len(sections)):
        start_pos = reduce(lambda x, y: x + len(y), sections[:x], 0)
        end_pos = start_pos + len(sections[x])
        slices.append((slice(start_pos, end_pos), sections[x][0]))
    slices.append((slice(len(optcodes), None), ' '))
    return slices


def diff(content_from, content_to):
    diff_lines = list(difflib.Differ().compare(
        content_from.splitlines(), content_to.splitlines()))

    data_lines = []
    line_no = 0
    line_padding = len(str(len(diff_lines)))
    line_no_a = line_no_b = 0

    for line in diff_lines:
        if line.startswith('? '):
            slices = optcodes_to_slices(line)
            spans = map(lambda x: {
                'type': DIFF_OPTCODE_MAP[x[1]],
                'text': data_lines[-1]['slices'][0]['text'][x[0]]}, slices)
            line = {
                'type': DIFF_OPTCODE_MAP['^'], 'slices': spans,
                'number_a': data_lines[-1]['number_a'],
                'number_b': data_lines[-1]['number_b']}
            data_lines[-1] = line
            if data_lines[-2]['type'] != DIFF_OPTCODE_MAP[' ']:
                data_lines[-2]['type'] = DIFF_OPTCODE_MAP['^']

        else:
            if line[:1] == '+':
                line_no_a += 1; number_a = line_no_a; number_b = ''
            elif line[:1] == '-':
                line_no_b += 1; number_a = ''; number_b = line_no_b
            elif line[:1] == ' ':
                line_no_a += 1; line_no_b += 1
                number_a = line_no_a; number_b = line_no_b
            data_lines.append({
                'type': DIFF_OPTCODE_MAP[line[:1]],
                'number_a': unicode(number_a).rjust(line_padding),
                'number_b': unicode(number_b).rjust(line_padding),
                'slices': [{'type': DIFF_OPTCODE_MAP[' '], 'text': line[2:]}]})

        line_no += 1

    return data_lines



import optparse

from django.core import management
from django.utils.translation import ugettext as _

from svnlit import models


class Command(management.BaseCommand):
    help = _('Add a repository')
    args = _('<label> <root>')

    option_list = (
        optparse.make_option(
            '--uri', action='store', dest='uri', default='',
            help=_('Externally facing URI for the repository, if available.')),
        optparse.make_option(
            '--private', action='store_true', dest='private',
            help=_('If browsing the repository requires authentication.')),
        optparse.make_option(
            '--username', action='store', dest='username', default='',
            help=_('Username to use when connecting to the repository.')),
        optparse.make_option(
            '--password', action='store', dest='password', default='',
            help=_('Password to use when connecting to the repository.')),
    ) + management.BaseCommand.option_list

    def handle(self, *args, **options):
        if not len(args) == 2:
            raise management.CommandError(
                _('wrong number of arguments, label and root required'))

        label, root = args

        models.Repository.objects.create(
            label=label, root=root, uri=options['uri'],
            username=options['username'], password=options['password'],
            is_private=options['private'] or False)
        
        print _('Repository %s added') % label



from django.conf import settings


if not hasattr(settings, 'SVNLIT_CLIENT_TIMEOUT'):
    settings.SVNLIT_CLIENT_TIMEOUT = 20

if not hasattr(settings, 'SVNLIT_CACHE_TIMEOUT'):
    settings.SVNLIT_CACHE_TIMEOUT = 60 * 60 * 24

if not hasattr(settings, 'SVNLIT_SVN_CONFIG_PATH'):
    settings.SVNLIT_SVN_CONFIG_PATH = None

if not hasattr(settings, 'SVNLIT_CHANGESETS_PER_PAGE'):
    settings.SVNLIT_CHANGESETS_PER_PAGE = 50

if not hasattr(settings, 'SVNLIT_AUTO_SYNC'):
    settings.SVNLIT_AUTO_SYNC = True

if not hasattr(settings, 'SVNLIT_SYNC_INTERVAL'):
    settings.SVNLIT_SYNC_INTERVAL = 60



import posixpath
from cStringIO import StringIO

from django import http, shortcuts, template
from django.conf import settings
from django.core.servers import basehttp
from django.core.urlresolvers import reverse
from django.views.generic import list_detail
import pysvn

from svnlit import models, forms, decorators, exceptions, diff


def repository_list(request):
    """A view listing the available repositories."""
    repository_list = models.Repository.objects.all()

    if not request.user.is_authenticated():
        repository_list = repository_list.filter(is_private=False)

    return shortcuts.render_to_response(
        'svnlit/repository_list.html', locals(),
        context_instance=template.RequestContext(request))

repository_list = decorators.autosync_repositories(repository_list)


def changeset_list(request, repository_label):
    """A view listing the changesets for the given repository."""
    repository_lookup = {'label': repository_label}
    if not request.user.is_authenticated():
        repository_lookup['is_private'] = False
    repository = shortcuts.get_object_or_404(
        models.Repository, **repository_lookup)

    if request.GET.get('results', '').isdigit():
        results = int(request.GET.get('results'))
    else:
        results = settings.SVNLIT_CHANGESETS_PER_PAGE

    qs = http.QueryDict('', mutable=True)
    qs.update(request.GET)
    if 'page' in qs:
        del qs['page']
    qs = qs.urlencode()

    return list_detail.object_list(
        request, repository.changesets.all(), paginate_by=results,
        extra_context={'repository': repository, 'qs': qs},
        template_object_name='changeset')

changeset_list = decorators.autosync_repositories(changeset_list)


def changeset(request, repository_label, revision):
    """A view showing a given changeset."""
    repository_lookup = {'label': repository_label}
    if not request.user.is_authenticated():
        repository_lookup['is_private'] = False
    repository = shortcuts.get_object_or_404(
        models.Repository, **repository_lookup)
    if not revision:
        revision = repository.get_latest_revision()
    changeset = shortcuts.get_object_or_404(
        models.Changeset, repository=repository, revision=revision)

    return shortcuts.render_to_response(
        'svnlit/changeset.html', locals(),
        context_instance=template.RequestContext(request))


def node(request, repository_label, revision, path):
    """A view for displaying a given path at a given revision."""
    if request.method == 'POST':
        r = request.POST.get('revision', '').lower()
        if r.startswith('r'):
            r = r[1:]
        if r.isdigit():
            return http.HttpResponseRedirect(reverse(
                'svnlit_node_revision', args=(repository_label, r, path)))
    
    repository_lookup = {'label': repository_label}
    if not request.user.is_authenticated():
        repository_lookup['is_private'] = False
    repository = shortcuts.get_object_or_404(
        models.Repository, **repository_lookup)
    if not revision:
        revision = repository.get_latest_revision()
    changeset = shortcuts.get_object_or_404(
        models.Changeset, repository=repository, revision=revision)

    if not path:
        path = posixpath.sep

    try:
        node = repository.get_node(path, revision)
    except exceptions.InvalidNode:
        return shortcuts.render_to_response(
            'svnlit/node_invalid.html', locals(),
            context_instance=template.RequestContext(request))

    if node.is_directory():
        return shortcuts.render_to_response(
            'svnlit/node_directory.html', locals(),
            context_instance=template.RequestContext(request))
    else:
        return shortcuts.render_to_response(
            'svnlit/node_file.html', locals(),
            context_instance=template.RequestContext(request))

node = decorators.autosync_repositories(node)


def content(request, repository_label, content_id, path):
    """
    A view for fetching a node's content.

    This is for serving images and other binary content stored in the
    repository.
    """
    repository_lookup = {'label': repository_label}
    if not request.user.is_authenticated():
        repository_lookup['is_private'] = False
    repository = shortcuts.get_object_or_404(
        models.Repository, **repository_lookup)
    content = shortcuts.get_object_or_404(
        models.Content, pk=content_id)

    response = http.HttpResponse(
        basehttp.FileWrapper(StringIO(content.get_data())),
        content_type=content.get_mimetype())
    response['content-length'] = content.size
    return response


def node_diff(request, repository_label, from_revision, to_revision, path):
    """View a diff of two revisions at a node."""
    repository_lookup = {'label': repository_label}
    if not request.user.is_authenticated():
        repository_lookup['is_private'] = False
    repository = shortcuts.get_object_or_404(
        models.Repository, **repository_lookup)

    from_changeset = shortcuts.get_object_or_404(
        models.Changeset, repository=repository, revision=from_revision)
    to_changeset = shortcuts.get_object_or_404(
        models.Changeset, repository=repository, revision=to_revision)

    try:
        from_node = repository.get_node(path, from_revision)
    except exceptions.InvalidNode:
        revision = from_revision
        return shortcuts.render_to_response(
            'svnlit/node_invalid.html', locals(),
            context_instance=template.RequestContext(request))
    try:
        to_node = repository.get_node(path, to_revision)
    except exceptions.InvalidNode:
        revision = to_revision
        return shortcuts.render_to_response(
            'svnlit/node_invalid.html', locals(),
            context_instance=template.RequestContext(request))

    if not (from_node.is_file() and to_node.is_file()):
        raise http.Http404('Invalid node type for diff.')

    if from_node.content.is_binary() or to_node.content.is_binary():
        raise http.Http404('Cannot diff binary nodes.')

    form = forms.RepositoryDiffForm({
        'from_path': from_node.path, 'from_revision': from_node.revision,
        'to_path': to_node.path, 'to_revision': to_node.revision})

    diff_data = diff.diff(
        from_node.content.get_data(),
        to_node.content.get_data())

    return shortcuts.render_to_response(
        'svnlit/node_diff.html', locals(),
        context_instance=template.RequestContext(request))


def repository_diff(request, repository_label):
    """View a diff of any two nodes in the repository."""
    repository_lookup = {'label': repository_label}
    if not request.user.is_authenticated():
        repository_lookup['is_private'] = False
    repository = shortcuts.get_object_or_404(
        models.Repository, **repository_lookup)

    form = forms.RepositoryDiffForm()

    # data posted by the diff form
    if request.method == 'POST':
        form = forms.RepositoryDiffForm(request.POST)
        if form.is_valid(repository):
            from_node = form.cleaned_data['from_node']
            to_node = form.cleaned_data['to_node']

            if from_node.path == to_node.path:
                # if the path matches, redirect to the sugary url
                return http.HttpResponseRedirect(reverse(
                    'svnlit_node_diff', args=(
                     repository_label, from_node.revision,
                     to_node.revision, to_node.path)))
            else:
                # redirect to self with the nodes in the query string
                # to expose the copy+pastable url
                url = reverse(
                    'svnlit_repository_diff', args=(repository_label,))
                qs = 'from=%s&to=%s' % (
                    '%s:r%s' % (from_node.path, from_node.revision),
                    '%s:r%s' % (to_node.path, to_node.revision))
                url = '%s?%s' % (url, qs)
                return http.HttpResponseRedirect(url)

    # display the diff
    elif request.method == 'GET' and request.GET:
        if not (request.GET.get('from') and request.GET.get('to')):
            return http.HttpResponseRedirect(reverse(
                'svnlit_repository_diff', args=(repository_label,)))

        from_path, from_revision = (
            request.GET.get('from').split(':r', 1) + [''])[:2]
        to_path, to_revision = (
            request.GET.get('to').split(':r', 1) + [''])[:2]

        form = forms.RepositoryDiffForm({
                'from_path': from_path, 'from_revision': from_revision,
                'to_path': to_path, 'to_revision': to_revision})

        if form.is_valid(repository):
            from_node = form.cleaned_data['from_node']
            to_node = form.cleaned_data['to_node']

            diff_data = diff.diff(
                from_node.content.get_data(),
                to_node.content.get_data())

    return shortcuts.render_to_response(
        'svnlit/repository_diff.html', locals(),
        context_instance=template.RequestContext(request))



from datetime import datetime

from django import template
from django.utils.timesince import timesince as timesince_


register = template.Library()


def timesince(dt, arg=None):
    if not dt:
        return ''
    if arg:
        reference_dt = arg
    else:
        reference_dt = datetime.now()
    if dt > reference_dt:
        return "in %s" % timesince_(reference_dt, dt)
    else:
        return "%s ago" % timesince_(dt, reference_dt)

register.filter('timesince', timesince)



"""
Models for representing subversion repositories.
"""

import datetime
import posixpath
import mimetypes

from django.db import models
from django.core import exceptions
from django.utils import functional, safestring
from django.utils.translation import ugettext_lazy as _

from svnlit import managers, choices


class Repository(models.Model):
    """
    Meta data for a subversion repository.
    """
    uuid = models.CharField(max_length=128, editable=False)
    label = models.CharField(
        help_text=_('Unique label for the repository for use in the URL'),
        max_length=128, db_index=True, unique=True)
    root = models.CharField(
        help_text=_('Example: svn://example.com or file:///svn/'),
        max_length=512)
    uri = models.CharField(
        help_text=_('Externally facing URI for the repository, if available'),
        max_length=512, blank=True)
    is_private = models.BooleanField(default=False)

    username = models.CharField(max_length=512, blank=True)
    password = models.CharField(max_length=512, blank=True)

    last_synced = models.DateTimeField(
        default=functional.curry(datetime.datetime.fromtimestamp, 0),
        editable=False)

    objects = managers.RepositoryManager()

    class Meta:
        verbose_name_plural = _('repositories')
        ordering = ('label',)

    def __unicode__(self):
        return u'%s %s' % (self.label, self.root)

    def _get_login(self, realm, username, may_save):
        if not (self.username and self.password):
            raise exceptions.ImproperlyConfigured(_(
                'repository requires authentication, '
                'but no username and password available'))
        return (True, self.username, self.password, True)

    def sync(self):
        """
        Update the repository's changesets, and obtain the UUID if
        it's not set.
        """
        Repository.objects.sync(self)

    def sync_uuid(self):
        """
        Sync the repository UUID.
        """
        Repository.objects.sync_uuid(self)

    def sync_changesets(self):
        """
        Get all changes made to the repository since the
        `last_updated` datetime.
        """
        Repository.objects.sync_changesets(self)
        
    def get_latest_revision(self):
        """
        Get the latest revision of the repository.
        """
        revision = 0
        if self.changesets.count():
            revision = self.changesets.all()[0].revision
        return revision

    def get_node(self, path, revision=None):
        """
        Get a `svnlit.models.Node` object at the given
        path. Optionally specify a revision.
        """
        return Node.objects.get_or_sync(self, path, revision)
        

class Changeset(models.Model):
    """
    The meta data about a revision in a subversion repository.
    """
    repository = models.ForeignKey(Repository, related_name='changesets')
    date = models.DateTimeField()
    revision = models.PositiveIntegerField(db_index=True)
    author = models.CharField(max_length=512)
    message = models.TextField()

    class Meta:
        unique_together = (('repository', 'revision'),)
        ordering = ('-revision',)

    def __unicode__(self):
        return u'r%s' % self.revision

    def get_absolute_url(self):
        return ('svnlit_changeset', (self.repository.label, self.revision))
    get_absolute_url = models.permalink(get_absolute_url)

    def get_previous(self):
        """Get the previous changeset in the repository."""
        try:
            return self.repository.changesets.get(revision=self.revision - 1)
        except Changeset.DoesNotExist:
            return None

    def get_next(self):
        """Get the next changeset in the repository."""
        try:
            return self.repository.changesets.get(revision=self.revision + 1)
        except Changeset.DoesNotExist:
            return None


class Change(models.Model):
    """
    A changed path in a changeset, including the action taken.
    """
    changeset = models.ForeignKey(Changeset, related_name='changes')
    path = models.CharField(max_length=2048, db_index=True)
    action = models.CharField(max_length=1)

    copied_from_path = models.CharField(max_length=2048, null=True)
    copied_from_revision = models.PositiveIntegerField(null=True)

    class Meta:
        unique_together = (('changeset', 'path'),)
        ordering = ('changeset', 'path')

    def __unicode__(self):
        return u'%s %s' % (self.action, self.path)

    def _get_base_change(self):
        if hasattr(self, '_base_change'):
            return self._base_change
        if self.copied_from_revision is not None:
            self._base_change = Change.objects.get(
                changeset__repository=self.changeset.repository,
                revision=self.copied_from_revision)
            return self._base_change

    def is_addition(self):
        return self.action == 'A'

    def is_modification(self):
        return self.action == 'M'

    def is_deletion(self):
        return self.action == 'D'


class Node(models.Model):
    """
    The meta data for a path at a revision in a repository.

    Nodes can be understood as 'views' of a particular path in a
    repository at a particular revision number (a revision that may or
    may not have made changes at that path/revision). A node's actual
    content is stored in a separate model object, since the content
    may remain unchanged across a number of revisions at a particular
    path. The `get_last_changeset` method can be used to obtain the
    changeset and revision in which the node's path was last changed.

    This model largely reflects the information available through the
    subversion api. The field `cached` indicates when the data was
    retrieved from the api, and `cached_indirectly` indicates whether
    or not the node was generated from an api call for the node or
    from a related node (parent or one of its possible
    children). Indirectly cached nodes (which are usually nodes
    created as placeholders for heirarchical connections instead of
    through a direct api call) require another api call to collect the
    remaining missing information. Nodes can be optionally be included
    in a regular cleanup.
    """
    repository = models.ForeignKey(Repository, related_name='nodes')
    parent = models.ForeignKey('Node', related_name='children', null=True)
    path = models.CharField(max_length=2048, db_index=True)
    node_type = models.CharField(max_length=1)
    size = models.PositiveIntegerField(default=0)
    last_changed = models.DateTimeField(null=True)

    revision = models.PositiveIntegerField()
    cached = models.DateTimeField(default=datetime.datetime.now)
    cached_indirectly = models.BooleanField(default=True)

    content = models.ForeignKey('Content', related_name='nodes', null=True)

    objects = managers.NodeManager()

    class Meta:
        unique_together = (('repository', 'path', 'revision'),)
        ordering = ('node_type', 'path')

    def __unicode__(self):
        return u'%s@%s' % (self.path, self.revision)

    def iter_path(self):
        """
        Returns a generator that 'walks` up the node hierarchy,
        yielding each parent path until the root node is reached ('/').
        """
        path = self.path
        yield path
        while path != posixpath.sep:
            path = posixpath.split(path)[0]
            yield path

    def iter_path_basename(self):
        """
        Returns a generator that 'walks' up the node hierarchy,
        yielding a tuple of the path, and the basename of the path for
        each parent node until the root node is reached ('/').
        """
        for path in self.iter_path():
            basename = posixpath.basename(path)
            if not basename:
                basename = self.repository.label
            yield (path, basename)

    def get_last_changeset(self):
        """Get the latest `Changeset` object that affected this node."""
        c = self.repository.changesets.filter(
            date__lte=self.last_changed)#.exclude(revision=self.revision)
        if c.count():
            return c[0]
        else:
            return self.repository.changesets.get(date=self.last_changed)

    def get_absolute_url(self):
        repository = self.repository
        if self.revision != repository.get_latest_revision():
            return (
                'svnlit_node_revision', (
                repository.label, self.revision, self.path))
        else:
            return ('svnlit_node', (repository.label, self.path))
    get_absolute_url = models.permalink(get_absolute_url)

    def get_basename(self):
        """
        The basename of the node, either a file name or a directory
        name.
        """
        basename = posixpath.basename(self.path)
        return basename

    def is_directory(self):
        """Whether the node is a directory."""
        return self.node_type == choices.NODE_TYPE_DIR

    def is_file(self):
        """Whether the node is a file."""
        return self.node_type == choices.NODE_TYPE_FILE

    def is_root(self):
        """Whether the node is the root node ('/')."""
        return self.is_directory() and self.path == posixpath.sep

    def has_properties(self):
        """Whether the node has subversion properties set."""
        if self.properties.count():
            return True
        return False


class Property(models.Model):
    """
    A property that has been set on a node.
    """
    node = models.ForeignKey(Node, related_name='properties')
    key = models.CharField(max_length=512, db_index=True)
    value = models.TextField()

    class Meta:
        unique_together = (('node', 'key'),)
        verbose_name_plural = 'properties'

    def __unicode__(self):
        return '%s: %s' % (self.key, self.value)


class Content(models.Model):
    """
    The contents of a node at a revision.

    The data is base64 encoded in the database to allow storage of
    binary data. The `set_data` and `get_data` methods should be used
    to manipulate a node's data. `cached` indicates when the contents
    were retrieved from the api. Content objects can optionally be
    part of a regular cleanup.
    """
    repository = models.ForeignKey(Repository, related_name='content')
    path = models.CharField(max_length=2048)
    last_changed = models.DateTimeField()
    cached = models.DateTimeField(default=datetime.datetime.now)
    size = models.PositiveIntegerField(default=0)
    data = models.TextField()

    class Meta:
        unique_together = (('repository', 'path', 'last_changed'),)

    def __unicode__(self):
        return '%s@%s' % (self.path, self.get_last_changeset())

    def set_data(self, data):
        self.size = len(data)
        self.data = data.encode('base64')

    def get_data(self):
        if hasattr(self, '_decoded_data'):
            return self._decoded_data
        self._decoded_data = self.data.decode('base64')
        return self._decoded_data

    def get_last_changeset(self):
        """Get the changeset in which this content was committed."""
        return self.repository.changesets.get(date=self.last_changed)

    def get_mimetype(self):
        """
        Get the mimetype of the content. This is determined by the
        extension of the basename of the path. Defaults to
        application/octet-stream if the mimetype cannot be determined.
        """
        mtype = mimetypes.guess_type(self.path)[0]
        if mtype is None:
            return 'application/octet-stream'
        return mtype

    def get_maintype(self):
        """
        Get the maintype of the mimetype, i.e. 'image' in 'image/png'.
        """
        return self.get_mimetype().split('/')[0]

    def get_subtype(self):
        """
        Get the subtype of the mimetype, i.e. 'png' in 'image/png'.
        """
        return self.get_mimetype().split('/')[-1]

    def get_absolute_url(self):
        return ('svnlit_content', (
                self.repository.label, self.pk, self.get_basename()))
    get_absolute_url = models.permalink(get_absolute_url)

    def is_binary(self):
        """
        Whether or not the content is binary. This is determined in
        part by the mimetype, but if the mimetype is not available,
        then if the data cannot be decoded into ascii it will be
        presumed a binary format.
        """
        mtype = mimetypes.guess_type(self.path)[0]
        if mtype is None:
            try:
                self.get_data().decode('ascii')
            except UnicodeDecodeError:
                return True
            return False
        if not mtype.startswith('text'):
            return True
        return False

    def get_basename(self):
        """Get the basename of the node's full path (the filename)."""
        basename = posixpath.basename(self.path)
        return basename

    def get_data_display(self):
        """
        Get the content for display in text. Binary formats are just
        shown as '(binary)'. Plain text formats get run through the
        appropriate pygments lexer if the package is available.
        """
        if self.is_binary():
            return _('<pre>(binary)</pre>')

        txt = self.get_data()

        try:
            from pygments import highlight, lexers
            from pygments.formatters import HtmlFormatter
        except ImportError:
            return safestring.mark_safe('<pre>%s</pre>' % txt)

        try:
            lexer = lexers.guess_lexer(txt)
        except lexers.ClassNotFound:
            try:
                lexer = lexers.get_lexer_for_filename(self.get_basename())
            except:
                return safestring.mark_safe('<pre>%s</pre>' % txt)

        return safestring.mark_safe(
            highlight(txt, lexer, HtmlFormatter(linenos=True)))



from ez_setup import use_setuptools
use_setuptools()

from setuptools import setup, find_packages


setup(
    name='django-svnlit',
    version='0.1',
    license='BSD',

    keywords='django subversion svn browser',
    description="Django-based subversion browser.",
    url='http://pypi.python.org/pypi/django-svnlit',

    author='Tamas Kemenczy',
    author_email='tamas.kemenczy@gmail.com',

    install_requires=(
        'pysvn',
    ),

    classifiers=(
        'Development Status :: 2 - Pre-Alpha',
        'Environment :: Web Environment',
        'Framework :: Django',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
    ),

    zip_safe=False,
    include_package_data=True,
    packages=find_packages(exclude=('ez_setup', 'examples', 'tests')),
)



from django.core import management
from django.utils.translation import ugettext as _

from svnlit import models


class Command(management.BaseCommand):
    help = _('Get repository changes')
    args = _('<repository repository ...>')
    
    def handle(self, *args, **options):
        if args:
            try:
                rlist = map(
                    lambda r: models.Repository.objects.get(label=r), args)
            except models.Repository.DoesNotExist, error:
                raise management.CommandError(error)
        else:
            rlist = models.Repository.objects.all()
        
        for r in rlist:
            print _('Syncing %(label)s...') % {'label': r.label}
            r.sync()
